// Funge JIT compiler
// Copyright (c) 2004 Marek Materzok
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.


#include <iostream>

#include <llvm/Transforms/Utils/Cloning.h>                                      
#include <llvm/Module.h>
#include <llvm/Instructions.h>
#include <llvm/Type.h>

#include "BflibInlinePass.hpp"
#include "Bflib.hpp"

namespace {
	llvm::RegisterOpt<BflibInlinePass> X("bflib_inline", "Inline");
};

bool
BflibInlinePass::doInitialization(llvm::Module &m)
{
	inlinable.clear();
	inlinable.insert(bflib::input_int);
	inlinable.insert(bflib::output_int);
	inlinable.insert(bflib::input_char);
	inlinable.insert(bflib::output_char);
	inlinable.insert(bflib::clear_stack);
	return true;
}

bool
BflibInlinePass::runOnFunction(llvm::Function &f)
{
	for (llvm::Function::iterator bbi = f.begin(), bbe = f.end(); bbi != bbe; ++bbi)
	{
		for (llvm::BasicBlock::iterator ii = bbi->begin(), ie = bbi->end(); ii != ie; )
		{
			if (llvm::CallInst *ci = llvm::dyn_cast<llvm::CallInst>(ii))
			{
				++ii;
				if (inlinable.count(ci->getCalledFunction()))
				{
					llvm::InlineFunction(ci);
					bbi = ii->getParent();
					ie = bbi->end();
				}
			}
			else ++ii;
		}
	}
	return true;
}

bool
BflibInlinePass::doFinalization(llvm::Module &m)
{
	return true;
}

