// Funge JIT compiler
// Copyright (c) 2004 Marek Materzok
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.


#include <iostream>

#include <llvm/Module.h>
#include <llvm/Instructions.h>
#include <llvm/Constants.h>
#include <llvm/Type.h>

#include "BflibLowerVectorPass.hpp"
#include "Bflib.hpp"

namespace {
	llvm::RegisterOpt<BflibLowerVectorPass> X("bflib_lowervector", "Lower vector");
};

bool
BflibLowerVectorPass::doInitialization(llvm::Module &m)
{
	return true;
}

bool
BflibLowerVectorPass::runOnBasicBlock(llvm::BasicBlock &bb)
{
	for (llvm::BasicBlock::iterator ii = bb.begin(), ie = bb.end(); ii != ie; )
	{
		if (llvm::CallInst *ci = llvm::dyn_cast<llvm::CallInst>(ii))
		{
			++ii;
			if (ci->getCalledFunction() == bflib::push_vector)
				lowerPushVector(ci);
			else if (ci->getCalledFunction() == bflib::pop_vector)
				lowerPopVector(ci);
		}
		else ++ii;
	}
	return true;
}

void
BflibLowerVectorPass::lowerPushVector(llvm::CallInst *ci)
{
	llvm::BasicBlock *cb = ci->getParent();
	llvm::Value *vector = ci->getOperand(1);
	for (int i = 0; i < bflib_fspace->ndims; i++)
	{
		std::vector<llvm::Value *> indices;
		indices.push_back(llvm::ConstantSInt::get(llvm::Type::IntTy, 0));
		indices.push_back(llvm::ConstantUInt::get(llvm::Type::UIntTy, 0));
		indices.push_back(llvm::ConstantSInt::get(llvm::Type::IntTy, i));
		llvm::Value *dim = new llvm::GetElementPtrInst(vector, indices, "", ci);
		dim = new llvm::LoadInst(dim, "", ci);
		new llvm::CallInst(bflib::push_int, dim, "", ci);
	}
	cb->getInstList().erase(ci);
}

void
BflibLowerVectorPass::lowerPopVector(llvm::CallInst *ci)
{
	llvm::BasicBlock *cb = ci->getParent();
	llvm::Value *vector = ci->getOperand(1);
	for (int i = 2; i >= bflib_fspace->ndims; i--)
	{
		std::vector<llvm::Value *> indices;
		indices.push_back(llvm::ConstantSInt::get(llvm::Type::IntTy, 0));
		indices.push_back(llvm::ConstantUInt::get(llvm::Type::UIntTy, 0));
		indices.push_back(llvm::ConstantSInt::get(llvm::Type::IntTy, i));
		llvm::Value *dim = new llvm::GetElementPtrInst(vector, indices, "", ci);
		new llvm::StoreInst(llvm::ConstantSInt::get(llvm::Type::IntTy, 0), dim, ci);
	}
	for (int i = bflib_fspace->ndims - 1; i >= 0; i--)
	{
		std::vector<llvm::Value *> indices;
		indices.push_back(llvm::ConstantSInt::get(llvm::Type::IntTy, 0));
		indices.push_back(llvm::ConstantUInt::get(llvm::Type::UIntTy, 0));
		indices.push_back(llvm::ConstantSInt::get(llvm::Type::IntTy, i));
		llvm::Value *dim = new llvm::GetElementPtrInst(vector, indices, "", ci);
		llvm::Value *v = new llvm::CallInst(bflib::pop_int, "", ci);
		new llvm::StoreInst(v, dim, ci);
	}
	cb->getInstList().erase(ci);
}

bool
BflibLowerVectorPass::doFinalization(llvm::Module &m)
{
	return true;
}

