// Funge JIT compiler
// Copyright (c) 2004 Marek Materzok
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.


#include <iostream>

#include <llvm/Module.h>
#include <llvm/Instructions.h>
#include <llvm/Constants.h>
#include <llvm/Type.h>

#include "BflibSimplifyStackPass.hpp"
#include "BflibLowerVectorPass.hpp"
#include "Bflib.hpp"

namespace {
	llvm::RegisterOpt<BflibSimplifyStackPass> X("bflib_simplifystack", "Simplify stack");
};

void 
BflibSimplifyStackPass::getAnalysisUsage(llvm::AnalysisUsage &info) const
{
	info.addRequired<BflibLowerVectorPass>();
	info.addPreserved<BflibLowerVectorPass>();
}

bool
BflibSimplifyStackPass::doInitialization(llvm::Module &m)
{
	return true;
}

bool
BflibSimplifyStackPass::runOnBasicBlock(llvm::BasicBlock &bb)
{
	std::vector<llvm::CallInst *> pushes;
	for (llvm::BasicBlock::iterator ii = bb.begin(), ie = bb.end(); ii != ie; )
	{
		if (llvm::CallInst *ci = llvm::dyn_cast<llvm::CallInst>(ii))
		{
			++ii;
			llvm::BasicBlock *cb = ci->getParent();
			llvm::Function *cf = ci->getCalledFunction();
			// TODO: add support for stack stack
			if (cf == bflib::push_int)
			{
				pushes.push_back(ci);
			}
			else if (cf == bflib::pop_int)
			{
				if (pushes.empty()) continue;
				llvm::CallInst *pci = pushes.back();
				pushes.pop_back();
				llvm::Value *pv = pci->getOperand(1);
				ci->replaceAllUsesWith(pv);
				cb->getInstList().erase(pci);
				cb->getInstList().erase(ci);
			}
			else if (cf == bflib::clear_stack)
			{
				while (!pushes.empty())
				{
					llvm::CallInst *pci = pushes.back();
					pushes.pop_back();
					cb->getInstList().erase(pci);
				}
			}
			else if (cf == bflib::push_int_many)
			{
				break; // TODO: ensure, that there aren't any push_many here
			}
			else if (cf == bflib::begin_block || cf == bflib::end_block || cf == bflib::stack_transfer)
			{
				// Stack-stack operations are currently unanalyzable
				addPushMany(pushes, ci);
			}
		}
		else ++ii;
	}
	addPushMany(pushes, bb.getTerminator());
	return true;
}

bool
BflibSimplifyStackPass::doFinalization(llvm::Module &m)
{
	return true;
}

void 
BflibSimplifyStackPass::addPushMany(std::vector<llvm::CallInst*> &list, llvm::Instruction *ins)
{
	if (list.size() <= 1)
	{
		list.clear();
		return;
	}
	llvm::Value *values = new llvm::CallInst(bflib::push_int_many, 
			llvm::ConstantSInt::get(llvm::Type::IntTy, list.size()), "", ins);
	std::vector<llvm::Value*> idx;
	idx.push_back(llvm::ConstantSInt::get(llvm::Type::IntTy, 1));
	for (std::vector<llvm::CallInst*>::iterator i = list.begin();
			i != list.end(); ++i)
	{
		llvm::BasicBlock *cb = (*i)->getParent();
		new llvm::StoreInst((*i)->getOperand(1), values, ins);
		values = new llvm::GetElementPtrInst(values, idx, "", ins);
		(*i)->getOperand(1);
		cb->getInstList().erase(*i);
	}
	list.clear();
}


