// Funge JIT compiler
// Copyright (c) 2004 Marek Materzok
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.

#ifndef __bflib_h
#define __bflib_h

#define DIMENSIONS 3
#define HASHTABLE_2SIZE 20
#define HASHTABLE_SIZE (1 << HASHTABLE_2SIZE)
#define MAXFUN 256
#define CELLPOOL_NUM 2048

#ifdef __cplusplus
extern "C" {
#endif

struct Stack
{
	int size;
	int ptr;
	int elsize;
	void *mem;
};

struct Vector
{
	int v[DIMENSIONS];
};

struct Thread
{
	struct Vector ip;
	struct Vector delta;
	struct Vector storageoffset;
};

struct FunStart
{
	int id;
	struct Vector delta;
	struct FunStart *next;
};

struct Cell
{
	int value;
	struct FunStart *fstart;
	struct Vector loc;
	struct Cell *next;
};

struct FungeSpace
{
	struct Cell *cells[HASHTABLE_SIZE];
	int min[DIMENSIONS];
	int max[DIMENSIONS];
	int ndims;
	struct Cell *cellpool;
};

extern struct Stack *bflib_stackstack;
extern struct Stack *bflib_topstack;
extern struct Thread *bflib_thread;
extern struct FungeSpace *bflib_fspace;
extern void (**bflib_functions)(void);

void stack_init(struct Stack *stack, int elsize);
void stack_clear(struct Stack *stack);
void stack_enlarge(struct Stack *stack, int nsize);
#define stack_top(stack) ((stack)->mem + (((stack)->ptr - 1) * (stack)->elsize))
#define stack_push(stack) ((stack)->ptr++, ((stack)->ptr >= (stack)->size) ? stack_enlarge((stack), (stack)->size * 2) : 0, stack_top((stack)))
#define stack_push_unchecked(stack) ((stack)->ptr++, stack_top((stack)))
#define stack_pop(stack) (stack)->ptr--;

struct Cell *bflib_get_cell(struct Vector *loc);

int bflib_vector_equal(struct Vector *v1, struct Vector *v2);
void bflib_vector_add(struct Vector *lhs, struct Vector *rhs);
void bflib_vector_wrapadd(struct Vector *lhs, struct Vector *rhs);

void bflib_init();

#ifdef __cplusplus
};
#endif

#endif

