// Funge JIT compiler
// Copyright (c) 2004 Marek Materzok
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.


#include <stdlib.h>
#include <string.h>

#include "bflib.h"

struct Stack *bflib_stackstack;
struct Stack *bflib_topstack;
struct Thread *bflib_thread;
struct FungeSpace *bflib_fspace;
void (**bflib_functions)(void);

void stack_init(struct Stack *stack, int elsize)
{
	stack->size = 32;
	stack->ptr = 0;
	stack->elsize = elsize;
	stack->mem = malloc(elsize * stack->size);
}

void stack_clear(struct Stack *stack)
{
	stack->size = 0;
	stack->ptr = 0;
	if (stack->mem) free(stack->mem);
	stack->mem = 0;
}

void stack_enlarge(struct Stack *stack, int nsize)
{
	if (nsize < stack->size) return;
	void *nmem = malloc(stack->elsize * nsize);
	memcpy(nmem, stack->mem, stack->elsize * stack->size);
	free(stack->mem);
	stack->mem = nmem;
	stack->size = nsize;
}

struct Cell *bflib_alloc_cell()
{
	struct Cell *ret;
	if (!bflib_fspace->cellpool)
	{
		bflib_fspace->cellpool = malloc(sizeof(struct Cell) * CELLPOOL_NUM);
		for (int n = 0; n < CELLPOOL_NUM - 1; n++)
		{
			bflib_fspace->cellpool[n].next = &bflib_fspace->cellpool[n+1];
		}
		bflib_fspace->cellpool[CELLPOOL_NUM - 1].next = 0;
	}
	ret = bflib_fspace->cellpool;
	bflib_fspace->cellpool = ret->next;
	ret->next = 0;
	return ret;
}

struct Cell *bflib_get_cell(struct Vector *loc)
{
	int hash;
	{
		int a = loc->v[2], b = loc->v[1], c = loc->v[0];
		a=a-b; a=a-c; a=a^(c>>13); 
		b=b-c; b=b-a; b=b^(a<<8);  
		c=c-a; c=c-b; c=c^(b>>13); 
		a=a-b; a=a-c; a=a^(c>>12); 
		b=b-c; b=b-a; b=b^(a<<16); 
		c=c-a; c=c-b; c=c^(b>>5);  
		a=a-b; a=a-c; a=a^(c>>3);  
		b=b-c; b=b-a; b=b^(a<<10); 
		c=c-a; c=c-b; c=c^(b>>15); 
		hash = c;
	}	

	int idx = hash & (HASHTABLE_SIZE - 1);
	struct Cell *ret;
	for (ret = bflib_fspace->cells[idx]; ret; 
			ret = ret->next)
	{
		if (bflib_vector_equal(loc, &ret->loc)) break;
	}
	if (!ret)
	{
		for (int i = 0; i < DIMENSIONS; i++)
		{
			if (loc->v[i] < bflib_fspace->min[i])
				bflib_fspace->min[i] = loc->v[i];
			if (loc->v[i] > bflib_fspace->max[i])
				bflib_fspace->max[i] = loc->v[i];
		}
		ret = bflib_alloc_cell();
		ret->value = ' ';
		ret->fstart = 0;
		ret->loc = *loc;
		ret->next = bflib_fspace->cells[idx];
		bflib_fspace->cells[idx] = ret;
	}
	return ret;
}

int bflib_vector_equal(struct Vector *v1, struct Vector *v2)
{
	for (int x = 0; x < DIMENSIONS; x++)
	{
		if (v1->v[x] != v2->v[x]) return 0;
	}
	return 1;
}

void bflib_vector_add(struct Vector *lhs, struct Vector *rhs)
{
	for (int x = 0; x < DIMENSIONS; x++)
	{
		lhs->v[x] += rhs->v[x];
	}
}

void bflib_vector_wrapadd(struct Vector *lhs, struct Vector *rhs)
{
	bflib_vector_add(lhs, rhs);
	int outbound = 0;
	for (int i = 0; i < DIMENSIONS; i++)
	{
		if (lhs->v[i] < bflib_fspace->min[i]
				|| lhs->v[i] > bflib_fspace->max[i])
		{
			outbound = 1;
			break;
		}
	}
	if (!outbound) return;

	unsigned int mtimes = ~0UL, bmt;
	for (int i = 0; i < DIMENSIONS; i++)
	{
		if (rhs->v[i] < 0)
		{
			bmt = (bflib_fspace->max[i] - lhs->v[i]) / (-rhs->v[i]);
			if (bmt < mtimes) mtimes = bmt;
		}
		else if (rhs->v[i] > 0)
		{
			bmt = (lhs->v[i] - bflib_fspace->min[i]) / rhs->v[i];
			if (bmt < mtimes) mtimes = bmt;
		}
	}
	for (int i = 0; i < DIMENSIONS; i++)
	{
		lhs->v[i] -= mtimes * rhs->v[i];
	}
}

