// Funge JIT compiler
// Copyright (c) 2004 Marek Materzok
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "bflib.h"

void bflib_push_int(int val)
{
	int *pval = (int*)stack_push(bflib_topstack);
	*pval = val;
}

int *bflib_push_int_many(int n)
{
	if (bflib_topstack->ptr + n >= bflib_topstack->size)
		stack_enlarge(bflib_topstack, bflib_topstack->size * 2 + n);

	bflib_topstack->ptr += n;
	int *pval = (int*)stack_top(bflib_topstack);
	return pval - n + 1;
}

int bflib_pop_int()
{
	if (bflib_topstack->ptr <= 0) return 0;
	int *pval = (int*)stack_top(bflib_topstack);
	int ret = *pval;
	stack_pop(bflib_topstack);
	return ret;
}

int bflib_input_int()
{
	int i;
	scanf("%ld", &i);
	return i;
}

void bflib_output_int(int i)
{
	printf("%ld ", i);
}

int bflib_input_char()
{
	return fgetc(stdin);
}

void bflib_output_char(int c)
{
	fputc(c, stdout);
}

void bflib_push_vector(struct Vector *pv)
{
	int *p = bflib_push_int_many(bflib_fspace->ndims);
	for (int i = 0; i < bflib_fspace->ndims; i++)
		p[i] = pv->v[i];
}

void bflib_pop_vector(struct Vector *pv)
{
	pv->v[2] = bflib_fspace->ndims >= 3 ? bflib_pop_int() : 0;
	pv->v[1] = bflib_fspace->ndims >= 2 ? bflib_pop_int() : 0;
	pv->v[0] = bflib_pop_int();
}

int bflib_fungespace_get(struct Vector *pv)
{
	struct Vector v = *pv;
	bflib_vector_add(&v, &bflib_thread->storageoffset);
	struct Cell *ce = bflib_get_cell(&v);
	return ce->value;
}

void bflib_fungespace_put(struct Vector *pv, int val)
{
	struct Vector v = *pv;
	bflib_vector_add(&v, &bflib_thread->storageoffset);
	struct Cell *ce = bflib_get_cell(&v);
	ce->value = val;
}

int bflib_fetch_char()
{
	struct Vector v = bflib_thread->ip;
	bflib_vector_add(&v, &bflib_thread->delta);
	struct Cell *ce = bflib_get_cell(&bflib_thread->ip);
	return ce->value;
}

void bflib_store_char(int c)
{
	struct Vector v = bflib_thread->ip;
	bflib_vector_add(&v, &bflib_thread->delta);
	struct Cell *ce = bflib_get_cell(&bflib_thread->ip);
	ce->value = c;
}

void bflib_clear_stack()
{
	bflib_topstack->ptr = 0;
}

void bflib_go_away()
{
	int r = rand() % (bflib_fspace->ndims * 2);
	bflib_thread->delta.v[0] = 0;
	bflib_thread->delta.v[1] = 0;
	bflib_thread->delta.v[2] = 0;
	switch (r)
	{
		case 0:
			bflib_thread->delta.v[0] = 1;
			break;
		case 1:
			bflib_thread->delta.v[0] = -1;
			break;
		case 2:
			bflib_thread->delta.v[1] = 1;
			break;
		case 3:
			bflib_thread->delta.v[1] = -1;
			break;
		case 4:
			bflib_thread->delta.v[2] = 1;
			break;
		case 5:
			bflib_thread->delta.v[2] = -1;
			break;
	}
	bflib_vector_wrapadd(&bflib_thread->ip, &bflib_thread->delta);
}

void bflib_begin_block()
{
	int n = bflib_pop_int();
	int mn = n < 0 ? -n : 0;
	if (mn) n = 0;
	for (int i = 0; i < mn; i++)
	{
		bflib_push_int(0);
	}
	int elements[n];
	for (int i = 0; i < n; i++)
	{
		elements[i] = bflib_pop_int();
	}
	bflib_push_vector(&bflib_thread->storageoffset);
	bflib_topstack = (struct Stack*)stack_push(bflib_stackstack);
	stack_init(bflib_topstack, sizeof(int));
	for (int i = n - 1; i >= 0; i--)
	{
		bflib_push_int(elements[i]);
	}
	bflib_thread->storageoffset = bflib_thread->ip;
	bflib_vector_add(&bflib_thread->storageoffset, &bflib_thread->delta);
	bflib_vector_wrapadd(&bflib_thread->ip, &bflib_thread->delta);
}

void bflib_end_block()
{
	if (bflib_stackstack->ptr == 0)
	{
		for (int i = 0; i < DIMENSIONS; i++)
			bflib_thread->delta.v[i] = bflib_thread->delta.v[i];
		bflib_vector_wrapadd(&bflib_thread->ip, &bflib_thread->delta);
		return;
	}
	int n = bflib_pop_int();
	int mn = n < 0 ? -n : 0;
	if (mn) n = 0;
	int elements[n];
	for (int i = 0; i < n; i++)
	{
		elements[i] = bflib_pop_int();
	}
	stack_clear(bflib_topstack);
	stack_pop(bflib_stackstack);
	bflib_topstack = (struct Stack*)stack_top(bflib_stackstack);
	bflib_pop_vector(&bflib_thread->storageoffset);
	for (int i = 0; i < mn; i++)
	{
		bflib_pop_int();
	}
	for (int i = n - 1; i >= 0; i--)
	{
		bflib_push_int(elements[i]);
	}
	bflib_vector_wrapadd(&bflib_thread->ip, &bflib_thread->delta);
}

void bflib_stack_transfer()
{
	if (bflib_stackstack->ptr == 0)
	{
		for (int i = 0; i < DIMENSIONS; i++)
			bflib_thread->delta.v[i] = bflib_thread->delta.v[i];
		bflib_vector_wrapadd(&bflib_thread->ip, &bflib_thread->delta);
		return;
	}
	struct Stack *soss = bflib_topstack - 1;
	struct Stack *toss = bflib_topstack;
	int n = bflib_pop_int();
	for (int i = 0; i < n; i++)
	{
		int val = *(int*)stack_top(soss);
		stack_pop(soss);
		*(int*)stack_push(toss);
	}
	for (int i = 0; i < -n; i++)
	{
		int val = *(int*)stack_top(toss);
		stack_pop(toss);
		*(int*)stack_push(soss);
	}
	bflib_vector_wrapadd(&bflib_thread->ip, &bflib_thread->delta);
}

int bfmain()
{
	while (1)
	{
		struct Cell *curcell = bflib_get_cell(&bflib_thread->ip);
		struct FunStart *fs;
		for (fs = curcell->fstart; fs; fs = fs->next)
		{
			if (bflib_vector_equal(&bflib_thread->delta, &fs->delta)) break;
		}
		if (!fs) return -1;
		bflib_functions[fs->id]();
	}
}

void bflib_init_stackstack()
{
	stack_init(bflib_stackstack, sizeof(struct Stack));
	bflib_topstack = (struct Stack*)stack_push(bflib_stackstack);
	stack_init(bflib_topstack, sizeof(int));
}

void bflib_init_thread(struct Thread *thr)
{
	for (int i = 0; i < DIMENSIONS; i++)
	{
		thr->ip.v[i] = 0;
		thr->delta.v[i] = 0;
		thr->storageoffset.v[i] = 0;
	}
	thr->delta.v[0] = 1;
}

void bflib_init_fungespace()
{
	for (int i = 0; i < DIMENSIONS; i++)
		bflib_fspace->min[i] = bflib_fspace->max[i] = 0;
	for (int i = 0; i < HASHTABLE_SIZE; i++)
		bflib_fspace->cells[i] = 0;
}

void bflib_init()
{
	bflib_stackstack = malloc(sizeof(struct Stack));
	bflib_thread = malloc(sizeof(struct Thread));
	bflib_fspace = malloc(sizeof(struct FungeSpace));
	bflib_functions = malloc(sizeof(void (*)())*MAXFUN);
	bflib_init_stackstack();
	bflib_init_thread(bflib_thread);
	bflib_init_fungespace();
}


